import 'package:data_db/data_db.dart';
import 'dart:io';
import 'package:file_utils/file_utils.dart';
import 'package:strings/strings.dart';

bool gernerateTableProto3(Dataset dataset, bool schemaInName, {String? path}) {
  var fullTableName = getFullTableName(dataset, schemaInName);

  var columns = deserializeDataset(dataset);
  var content = getStart(fullTableName)
      .add(getService(fullTableName))
      .add(getRequest(columns))
      .add(getResponse(fullTableName))
      .add(insert(fullTableName))
      .add(update(fullTableName))
      .add(delete(fullTableName))
      .add(select(fullTableName))
      .add(dataClass(columns, fullTableName))
      .add(selectableColumns(columns))
      .add(orderBy());

  path ??= 'services';
  writeInFile(path, getFileName(dataset, schemaInName), content);
  return true;
}

class _Column {
  late String name;
  late int typeId;
  dynamic? defaulValue;
  late bool isNullable;
  int? maxLength;
  int? precision;
  int? scale;
  bool isSerial = false;
  late bool isPk;
  late String fieldType;
  late String primitiveType;
  late String protoName;

  _Column(this.name, this.typeId, this.isNullable, this.isPk,
      [this.defaulValue, this.maxLength, this.precision, this.scale]) {
    protoName = camelize(name, true);
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
  }
  _Column.fromRecord(Map<String, DataField> record) {
    record.getValue('column_name').map((a) => a.tryGetValue((value) {
          if (value is String) {
            name = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('type_id').map((a) => a.tryGetValue((value) {
          if (value is int) {
            typeId = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('column_default').map((a) => a.tryGetValue((value) {
          if (value is String) {
            if (value.contains('nextval')) {
              isSerial = true;
            }
          }
          defaulValue = value;
          return true;
        }));
    record.getValue('is_nullable').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isNullable = value;
            return true;
          } else {
            return false;
          }
        }));
    record
        .getValue('character_maximum_length')
        .map((a) => a.tryGetValue((value) {
              if (value is int) {
                maxLength = value;
                return true;
              } else {
                return false;
              }
            }));
    record.getValue('numeric_precision').map((a) => a.tryGetValue((value) {
          if (value is int) {
            precision = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('numeric_scale').map((a) => a.tryGetValue((value) {
          if (value is int) {
            scale = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('is_pk').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isPk = value;
            return true;
          } else {
            return false;
          }
        }));
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
    protoName = camelize(name, true);
  }

  String getFieldType() {
    if ((typeId == 16)) {
      return 'BooleanValue';
    } else if ((typeId == 1266) || (typeId == 1083)) {
      return 'StringValue';
    } else if ((typeId == 1184) || (typeId == 1114)) {
      return 'StringValue';
    } else if ((typeId == 1082)) {
      return 'StringValue';
    } else if ((typeId == 20) ||
        (typeId == 21) ||
        (typeId == 23) ||
        (typeId == 26)) {
      return 'IntegerValue';
    } else if ((typeId == 700) ||
        (typeId == 701) ||
        (typeId == 790) ||
        (typeId == 1700)) {
      return 'NumericValue';
    } else if ((typeId == 25) ||
        (typeId == 19) ||
        (typeId == 18) ||
        (typeId == 1043)) {
      return 'StringValue';
    } else {
      return 'StringValue';
    }
  }

  String getCreaateField() {
    var createFieldContent =
        '_' "$name = $fieldType(this, '$name', ${!isNullable}, $isPk";
    if (fieldType.contains('StringField')) {
      if ((maxLength != null)) {
        createFieldContent += ', $maxLength';
      } else {
        createFieldContent += ', -1';
      }
    }
    if (defaulValue != null) {
      if (fieldType.contains('String')) {
        createFieldContent += ", '$defaulValue'";
      } else if (fieldType == 'TimeField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('HH:mm:ss').parse(${(defaulValue as String).substring(0, 9)}')";
      } else if (fieldType == 'DateTimeField_WithDefault_PG') {
        if ((defaulValue as String)[21] == ':') {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss').parse(${(defaulValue as String).substring(0, 20)}')";
        } else {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss').parse(${(defaulValue as String).substring(0, 24)}')";
        }
      } else if (fieldType == 'DateField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('yyyy-MM-dd').parse(${(defaulValue as String).substring(0, 11)}')";
      } else {
        if (fieldType != 'SerialField_PG') {
          createFieldContent += ', $defaulValue';
        }
      }
    }
    createFieldContent += ');';
    return createFieldContent;
  }

  String getPrimitiveType() {
    if (fieldType.contains('Integer')) {
      return 'int32';
    } else if (fieldType.contains('String')) {
      return 'string';
    } else if (fieldType.contains('Numeric')) {
      return 'double';
    } else if (fieldType.contains('Boolean')) {
      return 'bool';
    } else if (fieldType.contains('Date')) {
      return 'string';
    } else if (fieldType.contains('Time')) {
      return 'string';
    } else if (fieldType.contains('Serial')) {
      return 'int32';
    } else {
      return 'string';
    }
  }
}

extension Snake on String {
  String add(String complement) {
    return '$this\n$complement';
  }

  String capitalize() {
    return '${this[0].toUpperCase()}${substring(1)}';
  }
}

List<_Column> deserializeDataset(Dataset dataset) {
  var list = <_Column>[];
  dataset.forEach((record) {
    list.add(_Column.fromRecord(record));
  });

  return list;
}

String getFullTableName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '$schema.$name';
  } else {
    return '$name';
  }
}

String getTableName(Dataset dataset) {
  var name = '';
  dataset.tryGetValue('table_name', (table_name) {
    name = (table_name as String).toLowerCase();
    return true;
  });

  return name;
}

String getSchemaName(Dataset dataset) {
  var name = '';
  dataset.tryGetValue('table_schema', (table_name) {
    name = (table_name as String).toLowerCase();
    return true;
  });

  return name;
}

String getFileName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset).toLowerCase();
  var name = getTableName(dataset).toLowerCase();

  if (schemaInName) {
    return '${schema}_$name.proto';
  } else {
    return '$name.proto';
  }
}

String getClassName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '${schema.capitalize()}${name.capitalize()}Service';
  } else {
    return '${name.capitalize()}Service';
  }
}

void writeInFile(String filePath, String fileName, String contents) {
  FileUtils.mkdir([filePath]);
  File(filePath + '/' + fileName).writeAsString(contents);
}

String getStart(String fullTableName) {
  var content = 'syntax = "proto3";'
      .add('')
      .add('option java_multiple_files = true;')
      .add('option java_package = "static_grpc.$fullTableName";')
      .add(
          'option java_outer_classname = "${fullTableName.capitalize()}Proto";')
      .add(
          'option objc_class_prefix = "${fullTableName.replaceAll('.', '').toUpperCase()}";')
      .add('')
      .add('package static_grpc.$fullTableName;')
      .add('')
      .add('message IntegerValue{')
      .add('    int32 value = 1;  ')
      .add('}')
      .add('message StringValue{')
      .add('    string value = 1;  ')
      .add('}')
      .add('message NumericValue{')
      .add('    double value = 1; ')
      .add('}')
      .add('message BooleanValue{')
      .add('    bool value = 1; ')
      .add('}');

  return content;
}

String getService(String fullTableNamse) {
  var cl = fullTableNamse.replaceAll('.', '').capitalize();
  var content = ''
      .add('service ${cl}Service {')
      .add('    rpc get(GetRequest) returns (GetResponse);')
      .add('    rpc select(SelectRequest) returns (SelectResponse);')
      .add('    rpc insert(InsertRequest) returns (InsertResponse);')
      .add('    rpc update(UpdateRequest) returns (UpdateResponse);')
      .add('    rpc delete(DeleteRequest) returns (DeleteResponse);')
      .add('}');
  return content;
}

String getRequest(List<_Column> columns) {
  var content = ''.add('message GetRequest{');
  var number = 0;
  columns.where((element) => element.isPk).forEach((element) {
    number++;
    content = content.add(
        '    ${element.primitiveType} ${camelize(element.name, true)} = $number;');
  });
  content = content.add('}');
  return content;
}

String getResponse(String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''
      .add('message GetResponse {')
      .add('    bool result = 1; ')
      .add('    oneof GetResponseResult ')
      .add('    {')
      .add('        $name data = 2;')
      .add('        string errorMessage = 3;')
      .add('    }')
      .add('}');
  return content;
}

String insert(String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''
      .add('message InsertRequest{')
      .add('    repeated ${name}ToInsert data = 1;')
      .add('    InsertOptions options = 2;')
      .add('}')
      .add('')
      .add('message InsertOptions{')
      .add('    bool inTransaction = 1;')
      .add('}')
      .add('message InsertResponse{')
      .add('    bool result = 1;')
      .add('    int32 rowsToInsert = 2;')
      .add('    int32 rowsInserted = 3;')
      .add('    repeated string errorMessage = 4;')
      .add('}');
  return content;
}

String update(String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''
      .add('message UpdateRequest{')
      .add('    repeated $name data = 1;')
      .add('    UpdateOptions options = 2;')
      .add('}')
      .add('')
      .add('message UpdateOptions{')
      .add('    bool inTransaction = 1;')
      .add('}')
      .add('')
      .add('message UpdateResponse{')
      .add('    bool result = 1;')
      .add('    int32 rowsToUpdate = 2;')
      .add('    int32 rowsUpdated = 3;')
      .add('    repeated string errorMessage = 4;')
      .add('}');
  return content;
}

String delete(String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''
      .add('message DeleteRequest{')
      .add('    repeated ${name}Pk data = 1;')
      .add('    DeleteOptions options = 2;')
      .add('}')
      .add('')
      .add('message DeleteOptions{')
      .add('    bool inTransaction = 1;')
      .add('}')
      .add('')
      .add('message DeleteResponse{')
      .add('    bool result = 1;')
      .add('    int32 rowsToDelete = 2;')
      .add('    int32 rowsDeleted = 3;')
      .add('    repeated string errorMessage = 4;')
      .add('}');
  return content;
}

String select(String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''
      .add('message SelectRequest{')
      .add('    repeated SelectableColumns columns = 1;')
      .add('    string where = 2;')
      .add('    repeated OrderByOperator orderBy = 3;')
      .add('    int32 limit = 4;')
      .add('    int32 offset = 5;')
      .add('}')
      .add('')
      .add('message SelectResponse{')
      .add('    bool result = 1;')
      .add('    int32 rowsSelected = 2;')
      .add('    int32 totalRowsCount = 3;')
      .add('    int32 limit = 4;')
      .add('    int32 offset = 5;')
      .add('    int32 nextOffset = 6;')
      .add('    repeated $name data = 7; ')
      .add('    repeated string errorMessage = 8;')
      .add('}');
  return content;
}

String dataClass(List<_Column> columns, String fullTableName) {
  var name = fullTableName.replaceAll('.', '').capitalize();
  var content = ''.add('message $name{');
  var number = 0;
  columns.forEach((element) {
    number++;
    content =
        content.add('    ${element.fieldType} ${element.protoName} = $number;');
  });
  content = content.add('}');

  content = content.add('').add('message ${name}Pk{');
  number = 0;
  columns.where((element) => element.isPk).forEach((element) {
    number++;
    content =
        content.add('    ${element.fieldType} ${element.protoName} = $number;');
  });
  content = content.add('}');

  content = content.add('').add('message ${name}ToInsert{');
  number = 0;
  columns.where((element) => !element.isSerial).forEach((element) {
    number++;
    content =
        content.add('    ${element.fieldType} ${element.protoName} = $number;');
  });
  content = content.add('}');

  return content;
}

String selectableColumns(List<_Column> columns) {
  var content = ''.add('enum SelectableColumns {').add('    allColumns = 0;');
  var number = 0;
  columns.forEach((element) {
    number++;
    if (element.protoName == 'name') {
      content = content.add('    name_ = $number;');
    } else {
      content = content.add('    ${element.protoName} = $number;');
    }
  });
  content = content.add('}');
  return content;
}

String orderBy() {
  var content = ''
      .add('enum OrderType {')
      .add('    asc = 0;')
      .add('    desc = 1;')
      .add('}')
      .add('')
      .add('message OrderByOperator{')
      .add('    SelectableColumns column = 1;')
      .add('    OrderType orderType = 2; ')
      .add('}');
  return content;
}
