import 'package:data_db/data_db.dart';
import 'dart:io';
import 'package:file_utils/file_utils.dart';
import 'package:strings/strings.dart';

bool gernerateTableService(Dataset dataset, bool schemaInName,
    {String? path, String? subPath}) {
  var fullTableName = getFullTableName(dataset, schemaInName);
  var tableName = getTableName(dataset);
  var ormClassName = getOrmClassName(dataset, schemaInName);

  var columns = deserializeDataset(dataset);
  var content = getImports(
    columns,
    getOrmFileName(dataset, schemaInName),
    fullTableName,
  )
      .add(getServiceClass(getClassName(dataset, schemaInName)))
      .add(getColumnsToSelect())
      .add(resolveColumns(columns))
      .add(executeGet())
      .add(tryToMaterialize(ormClassName))
      .add(executeInsert(ormClassName, tableName))
      .add(finalizeInsertResponse())
      .add(performAllInserts(ormClassName, tableName))
      .add(tryToInsert(ormClassName))
      .add(executeUpdate(ormClassName, tableName))
      .add(finalizeUpdateResponse())
      .add(performAllUpdates(ormClassName, tableName))
      .add(tryToUpdate(ormClassName))
      .add(executeDelete(ormClassName, tableName))
      .add(finalizeDeleteResponse())
      .add(performAllDeletes(ormClassName, tableName))
      .add(tryToDelete(ormClassName))
      .add(tryToSelect(ormClassName))
      .add(assignNextPage())
      .add(getOrderBy())
      .add(orderByColumn(columns))
      .add('');
  path ??= 'services';
  var dirPath = path;
  if (subPath != null) {
    dirPath =
        '$path/${getClassName(dataset, schemaInName).replaceFirst('Service', '').toLowerCase()}/$subPath';
  }

  if (!Directory(dirPath).existsSync()) {
    Directory(dirPath).createSync(recursive: true);
  }

  writeInFile(dirPath, getFileName(dataset, schemaInName), content);
  return true;
}

class _Column {
  late String name;
  late int typeId;
  dynamic? defaulValue;
  late bool isNullable;
  int? maxLength;
  int? precision;
  int? scale;
  bool isSerial = false;
  late bool isPk;
  late String fieldType;
  late String primitiveType;

  _Column(this.name, this.typeId, this.isNullable, this.isPk,
      [this.defaulValue, this.maxLength, this.precision, this.scale]) {
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
  }
  _Column.fromRecord(Map<String, DataField> record) {
    record.getValue('column_name').map((a) => a.tryGetValue((value) {
          if (value is String) {
            name = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('type_id').map((a) => a.tryGetValue((value) {
          if (value is int) {
            typeId = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('column_default').map((a) => a.tryGetValue((value) {
          if (value is String) {
            if (value.contains('nextval')) {
              isSerial = true;
            }
          }
          defaulValue = value;
          return true;
        }));
    record.getValue('is_nullable').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isNullable = value;
            return true;
          } else {
            return false;
          }
        }));
    record
        .getValue('character_maximum_length')
        .map((a) => a.tryGetValue((value) {
              if (value is int) {
                maxLength = value;
                return true;
              } else {
                return false;
              }
            }));
    record.getValue('numeric_precision').map((a) => a.tryGetValue((value) {
          if (value is int) {
            precision = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('numeric_scale').map((a) => a.tryGetValue((value) {
          if (value is int) {
            scale = value;
            return true;
          } else {
            return false;
          }
        }));
    record.getValue('is_pk').map((a) => a.tryGetValue((value) {
          if (value is bool) {
            isPk = value;
            return true;
          } else {
            return false;
          }
        }));
    fieldType = getFieldType();
    primitiveType = getPrimitiveType();
  }

  String getFieldType() {
    if ((typeId == 16)) {
      if (defaulValue != null) {
        return 'BooleanField_WithDefault_PG';
      } else {
        return 'BooleanField_PG';
      }
    } else if ((typeId == 1266) || (typeId == 1083)) {
      if (defaulValue != null) {
        return 'TimeField_WithDefault_PG';
      } else {
        return 'TimeField_PG';
      }
    } else if ((typeId == 1184) || (typeId == 1114)) {
      if (defaulValue != null) {
        return 'DateTimeField_WithDefault_PG';
      } else {
        return 'DateTimeField_PG';
      }
    } else if ((typeId == 1082)) {
      if (defaulValue != null) {
        return 'DateField_WithDefault_PG';
      } else {
        return 'DateField_PG';
      }
    } else if ((typeId == 20) ||
        (typeId == 21) ||
        (typeId == 23) ||
        (typeId == 26)) {
      if (defaulValue != null) {
        if (isSerial) {
          return 'SerialField_PG';
        } else {
          return 'IntegerField_WithDefault_PG';
        }
      } else {
        return 'IntegerField_PG';
      }
    } else if ((typeId == 700) ||
        (typeId == 701) ||
        (typeId == 790) ||
        (typeId == 1700)) {
      if (defaulValue != null) {
        return 'NumericField_WithDefault_PG';
      } else {
        return 'NumericField_PG';
      }
    } else if ((typeId == 25) ||
        (typeId == 19) ||
        (typeId == 18) ||
        (typeId == 1043)) {
      if (defaulValue != null) {
        return 'StringField_WithDefault_PG';
      } else {
        return 'StringField_PG';
      }
    } else {
      return 'DynamicField';
    }
  }

  String getCreaateField() {
    var createFieldContent =
        '_' "$name = $fieldType(this, '$name', ${!isNullable}, $isPk";
    if (fieldType.contains('StringField')) {
      if ((maxLength != null)) {
        createFieldContent += ', $maxLength';
      } else {
        createFieldContent += ', -1';
      }
    }
    if (defaulValue != null) {
      if (fieldType.contains('String')) {
        createFieldContent += ", '$defaulValue'";
      } else if (fieldType == 'TimeField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('HH:mm:ss').parse(${(defaulValue as String).substring(0, 9)}')";
      } else if (fieldType == 'DateTimeField_WithDefault_PG') {
        if ((defaulValue as String)[21] == ':') {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss').parse(${(defaulValue as String).substring(0, 20)}')";
        } else {
          createFieldContent +=
              ", DateFormat('yyyy-MM-dd HH:mm:ss').parse(${(defaulValue as String).substring(0, 24)}')";
        }
      } else if (fieldType == 'DateField_WithDefault_PG') {
        createFieldContent +=
            ", DateFormat('yyyy-MM-dd').parse(${(defaulValue as String).substring(0, 11)}')";
      } else {
        if (fieldType != 'SerialField_PG') {
          createFieldContent += ', $defaulValue';
        }
      }
    }
    createFieldContent += ');';
    return createFieldContent;
  }

  String getPrimitiveType() {
    if (fieldType.contains('Integer')) {
      return 'int';
    } else if (fieldType.contains('String')) {
      return 'String';
    } else if (fieldType.contains('Numeric')) {
      return 'double';
    } else if (fieldType.contains('Boolean')) {
      return 'bool';
    } else if (fieldType.contains('Date')) {
      return 'DateTime';
    } else if (fieldType.contains('Time')) {
      return 'DateTime';
    } else if (fieldType.contains('Seria')) {
      return 'int';
    } else {
      return 'dynamic';
    }
  }
}

extension Snake on String {
  String add(String complement) {
    return '$this\n$complement';
  }

  String capitalize() {
    return '${this[0].toUpperCase()}${substring(1)}';
  }
}

List<_Column> deserializeDataset(Dataset dataset) {
  var list = <_Column>[];
  dataset.forEach((record) {
    list.add(_Column.fromRecord(record));
  });

  return list;
}

String getFullTableName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '$schema.$name';
  } else {
    return '$name';
  }
}

String getTableName(Dataset dataset) {
  var name = '';
  dataset.tryGetValue('table_name', (table_name) {
    name = (table_name as String).toLowerCase();
    return true;
  });

  return name;
}

String getSchemaName(Dataset dataset) {
  var name = '';
  dataset.tryGetValue('table_schema', (table_name) {
    name = (table_name as String).toLowerCase();
    return true;
  });

  return name;
}

String getFileName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '${schema}_${name}_service.dart';
  } else {
    return '${name}_service.dart';
  }
}

String getClassName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '${schema.capitalize()}${name.capitalize()}Service';
  } else {
    return '${name.capitalize()}Service';
  }
}

String getOrmClassName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '${schema.capitalize()}_${name.capitalize()}_ORM';
  } else {
    return '${name.capitalize()}_ORM';
  }
}

String getOrmFileName(Dataset dataset, bool schemaInName) {
  var schema = getSchemaName(dataset);
  var name = getTableName(dataset);

  if (schemaInName) {
    return '${schema}_${name}_orm.dart';
  } else {
    return '${name}_orm.dart';
  }
}

void writeInFile(String filePath, String fileName, String contents) {
  FileUtils.mkdir([filePath]);
  File(filePath + '/' + fileName).writeAsString(contents);
}

String getImports(
    List<_Column> columns, String ormFileName, String fullTableName) {
  var content = "import '$ormFileName';"
      .add("import '../proto_generated/$fullTableName.pb.dart';")
      .add("import '../proto_generated/$fullTableName.pbgrpc.dart';")
      .add("import 'package:grpc/grpc.dart' as grpc;")
      .add("import 'package:static_postgres_orm/static_postgres_orm.dart';")
      .add("import 'dart:async';")
      .add("import 'package:postgres/postgres.dart';")
      .add("import 'package:static_grpc/static_grpc.dart';");

  return content;
}

String getServiceClass(String className) {
  var content = ''
      .add('class $className extends ${className}Base {')
      .add('  final PostgreSQLConnection pgConnection;')
      .add('  $className(this.pgConnection) : super();')
      .add('')
      .add('  @override')
      .add(
          '  Future<GetResponse> get(grpc.ServiceCall call, GetRequest request) async {')
      .add('    return await executeGet(request, pgConnection);')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  Future<InsertResponse> insert(')
      .add('      grpc.ServiceCall call, InsertRequest request) async {')
      .add('    return await executeInsert(')
      .add('      request.data, request.options.inTransaction, pgConnection);')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  Future<UpdateResponse> update(')
      .add('      grpc.ServiceCall call, UpdateRequest request) async {')
      .add('    return await executeUpdate(')
      .add(
          '        request.data, request.options.inTransaction, pgConnection);')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  Future<DeleteResponse> delete(')
      .add('      grpc.ServiceCall call, DeleteRequest request) async {')
      .add('    return await executeDelete(')
      .add(
          '        request.data, request.options.inTransaction, pgConnection);')
      .add('  }')
      .add('')
      .add('  @override')
      .add('  Future<SelectResponse> select(')
      .add('      grpc.ServiceCall call, SelectRequest request) async {')
      .add('    return tryToSelect(request, newConnection(pgConnection));')
      .add('  }')
      .add('}');
  return content;
}

String getColumnsToSelect() {
  var content = ''
      .add('Columns getColumnsToSelect(List<SelectableColumns> cols) {')
      .add('  var result = Columns();')
      .add('  cols.forEach((element) {')
      .add('    resolveColumns(result, element);')
      .add('  });')
      .add('  return result;')
      .add('}');

  return content;
}

String resolveColumns(List<_Column> columns) {
  var content = ''
      .add('void resolveColumns(Columns columns, SelectableColumns column) {')
      .add('  switch (column) {');

  columns.forEach((element) {
    content = content;
    if (camelize(element.name, true) == 'name') {
      content = content.add('    case SelectableColumns.name_:');
    } else {
      content = content
          .add('    case SelectableColumns.${camelize(element.name, true)}:');
    }

    content = content.add('      columns.${element.name};').add('      break;');
  });

  content = content.add('  }').add('}');

  return content;
}

String executeGet() {
  var content = ''
      .add('Future<GetResponse> executeGet(')
      .add('    GetRequest request, PostgreSQLConnection connection) async {')
      .add(
          '  return proto3JsonToJsonMap(request.toProto3Json()).fold((error) {')
      .add('    return GetResponse()..errorMessage = error;')
      .add('  }, (proto3Json) async {')
      .add('    return await tryToMaterialize(proto3Json, connection);')
      .add('  });')
      .add('}');

  return content;
}

String tryToMaterialize(String ormClassName) {
  var ormVar = camelize(ormClassName, true);

  var content = ''
      .add('Future<GetResponse> tryToMaterialize(')
      .add(
          '    Map<String, dynamic> proto3Json, PostgreSQLConnection connection) async {')
      .add('  var $ormVar = $ormClassName(newConnection(connection));')
      .add('')
      .add(
          '  return (await $ormVar.materializeFromProto3Json(proto3Json)).fold((l) {')
      .add('    return GetResponse()')
      .add('      ..result = false')
      .add('      ..errorMessage = l.toJson();')
      .add('  }, (r) {')
      .add('    return GetResponse()')
      .add('      ..result = true')
      .add('      ..mergeFromProto3Json($ormVar.recordToProto3Json());')
      .add('  });')
      .add('}');
  return content;
}

String executeInsert(String ormClassName, String tableName) {
  var content = ''
      .add(
          'Future<InsertResponse> executeInsert(List<${tableName.capitalize()}ToInsert> data,')
      .add('    bool inTransaction, PostgreSQLConnection connection) async {')
      .add('  var response = InsertResponse();')
      .add('  final nextCompleter = Completer.sync();')
      .add('  var pgconnection = newPGConnection(connection);')
      .add('  try {')
      .add('    if (inTransaction) {')
      .add('      await pgconnection.open();')
      .add('      await pgconnection.transaction((connection) async {')
      .add('        response.errorMessage.addAll(await performAllInserts(')
      .add(
          '            data, $ormClassName.fromConnection(connection), inTransaction));')
      .add('      });')
      .add('    } else {')
      .add('      response.errorMessage.addAll(await performAllInserts(')
      .add(
          '          data, $ormClassName(newConnection(pgconnection)), inTransaction));')
      .add('    }')
      .add('')
      .add('    nextCompleter.complete();')
      .add('  } catch (e) {')
      .add('    response.errorMessage.add(e.toString());')
      .add('    nextCompleter.complete();')
      .add('  } finally {')
      .add('    if (!pgconnection.isClosed) await pgconnection.close();')
      .add('  }')
      .add('')
      .add('  await nextCompleter.future;')
      .add('  finalizeInsertResponse(response, data.length, inTransaction);')
      .add('  return response;')
      .add('}');

  return content;
}

String finalizeInsertResponse() {
  var content = ''
      .add('void finalizeInsertResponse(')
      .add(
          '    InsertResponse response, int rowsToInsert, bool inTransaction) {')
      .add('  if (rowsToInsert <= 0) {')
      .add("    response.errorMessage.add('No record to insert');")
      .add('    response.rowsInserted = 0;')
      .add('  } else {')
      .add('    if (inTransaction && response.errorMessage.isNotEmpty) {')
      .add('      response.rowsInserted = 0;')
      .add('    } else {')
      .add(
          '      response.rowsInserted = rowsToInsert - response.errorMessage.length;')
      .add('    }')
      .add('  }')
      .add('  response.result = response.errorMessage.isEmpty;')
      .add('  response.rowsToInsert = rowsToInsert;')
      .add('}');

  return content;
}

String performAllInserts(
  String ormClassName,
  String tableName,
) {
  var content = ''
      .add('Future<List<String>> performAllInserts(')
      .add(
          '    List<${tableName.capitalize()}ToInsert> data, $ormClassName orm, bool inTransaction) async {')
      .add('  var result = <String>[];')
      .add('  for (var $tableName in data) {')
      .add('    var error = (await tryToInsert(')
      .add('        orm, $tableName.toProto3Json() as Map<String, dynamic>));')
      .add('    if (error.isNotEmpty) {')
      .add('      result.addAll(error);')
      .add('      if (inTransaction) {')
      .add('        break;')
      .add('      }')
      .add('    }')
      .add('  }')
      .add('  return result;')
      .add('}');

  return content;
}

String tryToInsert(String ormClassName) {
  var content = ''
      .add('Future<List<String>> tryToInsert(')
      .add('    $ormClassName orm, Map<String, dynamic> data) async {')
      .add('  try {')
      .add('    var result = <String>[];')
      .add('    (await orm.insertFromProto3Json(data))')
      .add('        .leftMap((l) => result.add(l.toJson()));')
      .add('    return result;')
      .add('  } catch (e) {')
      .add('    return [e.toString()];')
      .add('  }')
      .add('}');

  return content;
}

String executeUpdate(String ormClassName, String tableName) {
  var content = ''
      .add(
          'Future<UpdateResponse> executeUpdate(List<${tableName.capitalize()}> data,')
      .add('    bool inTransaction, PostgreSQLConnection connection) async {')
      .add('  var response = UpdateResponse();')
      .add('  final nextCompleter = Completer.sync();')
      .add('  var pgconnection = newPGConnection(connection);')
      .add('  try {')
      .add('    if (inTransaction) {')
      .add('      await pgconnection.open();')
      .add('      await pgconnection.transaction((connection) async {')
      .add('        response.errorMessage.addAll(await performAllUpdates(')
      .add(
          '            data, $ormClassName.fromConnection(connection), inTransaction));')
      .add('      });')
      .add('    } else {')
      .add('      response.errorMessage.addAll(await performAllUpdates(')
      .add(
          '          data, $ormClassName(newConnection(pgconnection)), inTransaction));')
      .add('    }')
      .add('')
      .add('    nextCompleter.complete();')
      .add('  } catch (e) {')
      .add('    response.errorMessage.add(e.toString());')
      .add('    nextCompleter.complete();')
      .add('  } finally {')
      .add('    if (!pgconnection.isClosed) await pgconnection.close();')
      .add('  }')
      .add('')
      .add('  await nextCompleter.future;')
      .add('  finalizeUpdateResponse(response, data.length, inTransaction);')
      .add('  return response;')
      .add('}');

  return content;
}

String finalizeUpdateResponse() {
  var content = ''
      .add('void finalizeUpdateResponse(')
      .add(
          '    UpdateResponse response, int rowsToUpdate, bool inTransaction) {')
      .add('  if (rowsToUpdate <= 0) {')
      .add("    response.errorMessage.add('No record to update');")
      .add('    response.rowsToUpdate = 0;')
      .add('  } else {')
      .add('    if (inTransaction && response.errorMessage.isNotEmpty) {')
      .add('      response.rowsToUpdate = 0;')
      .add('    } else {')
      .add(
          '      response.rowsUpdated = rowsToUpdate - response.errorMessage.length;')
      .add('    }')
      .add('  }')
      .add('  response.result = response.errorMessage.isEmpty;')
      .add('  response.rowsToUpdate = rowsToUpdate;')
      .add('}');

  return content;
}

String performAllUpdates(String ormClassName, String tableName) {
  var content = ''
      .add('Future<List<String>> performAllUpdates(')
      .add(
          '    List<${tableName.capitalize()}> data, $ormClassName orm, bool inTransaction) async {')
      .add('  var result = <String>[];')
      .add('  for (var record in data) {')
      .add('    var error =')
      .add(
          '        (await tryToUpdate(orm, record.toProto3Json() as Map<String, dynamic>));')
      .add('')
      .add('    if (error.isNotEmpty) {')
      .add('      result.addAll(error);')
      .add('      if (inTransaction) {')
      .add('        break;')
      .add('      }')
      .add('    }')
      .add('  }')
      .add('  return result;')
      .add('}');
  return content;
}

String tryToUpdate(String ormClassName) {
  var content = ''
      .add('Future<List<String>> tryToUpdate(')
      .add('    $ormClassName orm, Map<String, dynamic> data) async {')
      .add('  try {')
      .add('    var result = <String>[];')
      .add('    (await orm.updateFromProto3Json(data))')
      .add('        .leftMap((l) => result.add(l.toJson()));')
      .add('    return result;')
      .add('  } catch (e) {')
      .add('    return [e.toString()];')
      .add('  }')
      .add('}');

  return content;
}

String executeDelete(String ormClassName, String tableName) {
  var content = ''
      .add(
          'Future<DeleteResponse> executeDelete(List<${tableName.capitalize()}Pk> pkData,')
      .add('    bool inTransaction, PostgreSQLConnection connection) async {')
      .add('  var response = DeleteResponse();')
      .add('  final nextCompleter = Completer.sync();')
      .add('  var pgconnection = newPGConnection(connection);')
      .add('  try {')
      .add('    if (inTransaction) {')
      .add('      await pgconnection.open();')
      .add('      await pgconnection.transaction((connection) async {')
      .add('      response.errorMessage.addAll(await performAllDeletes(')
      .add(
          '          pkData, $ormClassName.fromConnection(connection), inTransaction));')
      .add('      });')
      .add('    } else {')
      .add('      response.errorMessage.addAll(await performAllDeletes(')
      .add(
          '          pkData, $ormClassName(newConnection(pgconnection)), inTransaction));')
      .add('    }')
      .add('')
      .add('    nextCompleter.complete();')
      .add('  } catch (e) {')
      .add('    response.errorMessage.add(e.toString());')
      .add('    nextCompleter.complete();')
      .add('  } finally {')
      .add('    if (!pgconnection.isClosed) await pgconnection.close();')
      .add('  }')
      .add('')
      .add('  await nextCompleter.future;')
      .add('  finalizeDeleteResponse(response, pkData.length, inTransaction);')
      .add('  return response;')
      .add('}');

  return content;
}

String finalizeDeleteResponse() {
  var content = ''
      .add('void finalizeDeleteResponse(')
      .add(
          '    DeleteResponse response, int rowsToDelete, bool inTransaction) {')
      .add('  if (rowsToDelete <= 0) {')
      .add("    response.errorMessage.add('No record to delete');")
      .add('    response.rowsToDelete = 0;')
      .add('  } else {')
      .add('    if (inTransaction && response.errorMessage.isNotEmpty) {')
      .add('      response.rowsToDelete = 0;')
      .add('    } else {')
      .add(
          '      response.rowsDeleted = rowsToDelete - response.errorMessage.length;')
      .add('    }')
      .add('  }')
      .add('  response.result = response.errorMessage.isEmpty;')
      .add('  response.rowsToDelete = rowsToDelete;')
      .add('}');

  return content;
}

String performAllDeletes(String ormClassName, String tableName) {
  var content = ''
      .add('Future<List<String>> performAllDeletes(')
      .add(
          '    List<${tableName.capitalize()}Pk> pkData, $ormClassName orm, bool inTransaction) async {')
      .add('  var result = <String>[];')
      .add('  for (var record in pkData) {')
      .add('    var error =')
      .add(
          '        (await tryToDelete(orm, record.toProto3Json() as Map<String, dynamic>));')
      .add('')
      .add('    if (error.isNotEmpty) {')
      .add('      result.addAll(error);')
      .add('      if (inTransaction) {')
      .add('         break;')
      .add('      }')
      .add('    }')
      .add('  }')
      .add('  return result;')
      .add('}');

  return content;
}

String tryToDelete(String ormClassName) {
  var content = ''
      .add('Future<List<String>> tryToDelete(')
      .add('    $ormClassName orm, Map<String, dynamic> data) async {')
      .add('  try {')
      .add('    var result = <String>[];')
      .add('    (await orm.deleteFromProto3Json(data))')
      .add('        .leftMap((l) => result.add(l.toJson()));')
      .add('    return result;')
      .add('  } catch (e) {')
      .add('    return [e.toString()];')
      .add('  }')
      .add('}');

  return content;
}

String tryToSelect(String ormClassName) {
  var content = ''
      .add('Future<SelectResponse> tryToSelect(')
      .add(
          '    SelectRequest request, Postgres_SqlConnection connection) async {')
      .add('  var response = SelectResponse();')
      .add('  var where = Where();')
      .add('  var limit = 20;')
      .add('  var offset = 0;')
      .add('')
      .add('  if (request.hasWhere()) {')
      .add('    where.loadFromJson(request.where);')
      .add('  }')
      .add('  if (request.hasLimit()) {')
      .add('    limit = request.limit;')
      .add('  }')
      .add('  if (request.hasOffset()) {')
      .add('    offset = request.offset;')
      .add('  }')
      .add('  if (request.hasWhere()) {')
      .add('    where.loadFromJson(request.where);')
      .add('  }')
      .add('  var ds = ((await $ormClassName(connection).select(')
      .add('          columns: getColumnsToSelect(request.columns),')
      .add('          where: where,')
      .add('          orderBy: getOrderBy(request.orderBy),')
      .add('          limit: limit,')
      .add('          offset: offset))')
      .add('      .toProto3Json());')
      .add('  response.mergeFromProto3Json(ds);')
      .add('')
      .add('  response.limit = limit;')
      .add('  response.offset = offset;')
      .add('  assignNextPage(response);')
      .add('  return response;')
      .add('}');

  return content;
}

String assignNextPage() {
  var content = ''
      .add('void assignNextPage(SelectResponse response) {')
      .add('  if (response.totalRowsCount > 0) {')
      .add('    int totalPages;')
      .add('    totalPages = response.totalRowsCount ~/ response.limit;')
      .add('    if ((response.totalRowsCount % response.limit) > 0) {')
      .add('      totalPages++;')
      .add('    }')
      .add('    if ((totalPages - 1) > response.offset) {')
      .add('      response.nextOffset = response.offset + 1;')
      .add('    }')
      .add('  }')
      .add('}');

  return content;
}

String getOrderBy() {
  var content = ''
      .add('OrderBy getOrderBy(List<OrderByOperator> operators) {')
      .add('  var result = OrderBy();')
      .add('  operators.forEach((element) {')
      .add('    if (element.column != SelectableColumns.allColumns) {')
      .add('      result.add(element.toOrdenator());')
      .add('    }')
      .add('  });')
      .add('  return result;')
      .add('}');
  return content;
}

String orderByColumn(List<_Column> columns) {
  var content = ''
      .add('extension OrderByColumn on OrderByOperator {')
      .add('  ColumnOrdenator toOrdenator() {')
      .add("    var name = '';")
      .add('    switch (column) {');
  columns.forEach((element) {
    if (camelize(element.name, true) == 'name') {
      content = content.add('      case SelectableColumns.name_:');
    } else {
      content = content
          .add('      case SelectableColumns.${camelize(element.name, true)}:');
    }
    content =
        content.add("        name = '${element.name}';").add('        break;');
  });
  content = content
      .add('   }')
      .add('    OrdType ordType;')
      .add('    if (orderType == OrderType.desc) {')
      .add('      ordType = OrdType.desc;')
      .add('    } else {')
      .add('      ordType = OrdType.asc;')
      .add('    }')
      .add('    return ColumnOrdenator(name, ordType);')
      .add('  }')
      .add('}');
  return content;
}
